/**
 * Copyright © 2005-2012 Akiban Technologies, Inc.  All rights reserved.
 * 
 * This program and the accompanying materials are made available
 * under the terms of the Eclipse Public License v1.0 which
 * accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program may also be available under different license terms.
 * For more information, see www.akiban.com or contact licensing@akiban.com.
 * 
 * Contributors:
 * Akiban Technologies, Inc.
 */

package com.persistit;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.AbstractQueuedSynchronizer;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import com.persistit.exception.PersistitInterruptedException;

/**
 * Base class for objects that need synchronization across threads. The methods
 * {@link #claim(boolean)} and {@link #register()} establish and release either
 * exclusive or non-exclusive access to the object. Numerous other status flags
 * are also managed by this class, such as {@link #isDirty()} and
 * {@link #isValid()}.
 * <p>
 * The implementation of this class uses a {@link AbstractQueuedSynchronizer}
 * and is similar to {@link ReentrantReadWriteLock}. The synchronization policy
 * is a non-strict fair policy which is necessary and sufficient to prevent
 * starvation on busy system.
 * 
 * See {@link Buffer}, {@link Tree} and {@link Volume}.
 * 
 * @author peter
 */
class SharedResource {

    /**
     * Default maximum time to wait for access to this resource. Methods throw
     * an InUseException when this time is exceeded.
     */
    public final static long DEFAULT_MAX_WAIT_TIME = 180000L;

    /**
     * Mask for count of Threads holding a reader or writer claim (0-32767)
     */
    final static int CLAIMED_MASK = 0x00007FFF;
    /**
     * Mask for count of Threads holding a writer claim (0 or 1)
     */
    final static int WRITER_MASK = 0x00008000;
    /**
     * Mask for field indicating the resource needs to be written
     */
    final static int DIRTY_MASK = 0x00010000;

    /**
     * Status field mask for valid bit. The bit is set if the contents of the
     * buffer accurately reflects the status of the page.
     */
    final static int VALID_MASK = 0x00020000;

    /**
     * Status field mask for a resource that is dirty but not required to be
     * written with any checkpoint.
     */
    final static int TEMPORARY_MASK = 0x00400000;

    /**
     * Status field mask indicating a resource has been touched. Used by
     * clock-based page replacement algorithm.
     */
    final static int TOUCHED_MASK = 0x08000000;

    /**
     * Mask for bit field indicating that resource (a Buffer) should not be
     * replaced. The buffer houses a Volume's head page.
     */
    final static int FIXED_MASK = 0x40000000;

    final static AtomicLong ACQUIRE_LOOPS = new AtomicLong();
    final static AtomicLong RELEASE_LOOPS = new AtomicLong();
    final static AtomicLong SET_BIT_LOOPS = new AtomicLong();
    final static AtomicLong CLEAR_BIT_LOOPS = new AtomicLong();

    /**
     * Extension of {@link AbstractQueuedSynchronizer} with Persistit semantics.
     */
    private static class Sync extends AbstractQueuedSynchronizer {
        private static final long serialVersionUID = 1L;

        @Override
        protected boolean tryAcquire(final int arg) {
            assert arg == 1;
            //
            // Implement non-strict fairness doctrine, as suggested in
            // download.oracle.com/javase/6/docs/api/java/util/concurrent/locks/AbstractQueuedSynchronizer.html
            //
            final Thread thisThread = Thread.currentThread();
            for (;;) {
                final Thread queuedThread = getFirstQueuedThread();
                if (queuedThread != null && queuedThread != thisThread && getExclusiveOwnerThread() != thisThread) {
                    return false;
                }
                final int state = getState();
                if (!isAvailable(state)) {
                    return false;
                } else if (compareAndSetState(state, (state | WRITER_MASK) + 1)) {
                    setExclusiveOwnerThread(thisThread);
                    return true;
                }
                ACQUIRE_LOOPS.incrementAndGet();
            }
        }

        @Override
        protected int tryAcquireShared(final int arg) {
            assert arg == 1;
            //
            // Implement non-strict fairness doctrine, as suggested in
            // download.oracle.com/javase/6/docs/api/java/util/concurrent/locks/AbstractQueuedSynchronizer.html
            //
            final Thread thisThread = Thread.currentThread();
            for (;;) {
                final Thread queuedThread = getFirstQueuedThread();
                if (queuedThread != null && queuedThread != thisThread && getExclusiveOwnerThread() != thisThread) {
                    return -1;
                }
                final int state = getState();
                if (!isAvailableShared(state)) {
                    return -1;
                } else if (compareAndSetState(state, state + 1)) {
                    return CLAIMED_MASK - (state & CLAIMED_MASK) - 1;
                }
                ACQUIRE_LOOPS.incrementAndGet();
            }
        }

        /**
         * Attempt to convert shared to exclusive. The caller must already have
         * acquired shared access. This method upgrades the state to exclusive,
         * but only if there is exactly one shared acquire.
         * 
         * TODO - prove that caller already a reader claim
         * 
         * @return
         */
        private boolean tryUpgrade() {
            for (;;) {
                final int state = getState();
                final Thread thisThread = Thread.currentThread();
                if ((state & CLAIMED_MASK) != 1 || ((state & WRITER_MASK) != 0)
                        && getExclusiveOwnerThread() != thisThread) {
                    return false;
                } else if (compareAndSetState(state, state | WRITER_MASK)) {
                    setExclusiveOwnerThread(thisThread);
                    return true;
                }
                ACQUIRE_LOOPS.incrementAndGet();
            }
        }

        @Override
        protected boolean tryRelease(final int arg) {
            return (releaseState(arg) & WRITER_MASK) == 0;
        }

        @Override
        protected boolean tryReleaseShared(final int arg) {
            return (releaseState(arg) & WRITER_MASK) == 0;
        }

        @Override
        protected boolean isHeldExclusively() {
            return (getState() & WRITER_MASK) != 0;
        }

        private boolean isAvailable(final int state) {
            return (state & CLAIMED_MASK) < CLAIMED_MASK
                    && ((state & CLAIMED_MASK) == 0 || getExclusiveOwnerThread() == Thread.currentThread());
        }

        private boolean isAvailableShared(final int state) {
            return (state & CLAIMED_MASK) < CLAIMED_MASK
                    && ((state & WRITER_MASK) == 0 || getExclusiveOwnerThread() == Thread.currentThread());
        }

        private int releaseState(final int count) {
            assert count == 0 || count == 1;
            for (;;) {
                final int state = getState();
                if ((state & CLAIMED_MASK) == 1) {
                    final int newState = (state - count) & ~WRITER_MASK;
                    // Do this first so that another thread setting
                    // a writer claim does not lose its copy.
                    setExclusiveOwnerThread(null);
                    if (compareAndSetState(state, newState)) {
                        return newState;
                    }
                } else if ((state & CLAIMED_MASK) > 1) {
                    if (count == 0) {
                        return state;
                    }
                    final int newState = state - 1;
                    if (compareAndSetState(state, newState)) {
                        return newState;
                    }
                } else {
                    throw new IllegalMonitorStateException("Unmatched attempt to release " + this);
                }
                RELEASE_LOOPS.incrementAndGet();
            }
        }

        // Visible to outer class
        private int state() {
            return getState();
        }

        private Thread writerThread() {
            return getExclusiveOwnerThread();
        }

        private boolean setBitsInState(final int mask) {
            for (;;) {
                final int state = getState();
                final int newState = state | mask;
                if (compareAndSetState(state, newState)) {
                    return state != newState;
                }
                SET_BIT_LOOPS.incrementAndGet();
            }
        }

        private boolean clearBitsInState(final int mask) {
            for (;;) {
                final int state = getState();
                final int newState = state & ~mask;
                if (compareAndSetState(state, newState)) {
                    return state != newState;
                }
                CLEAR_BIT_LOOPS.incrementAndGet();
            }
        }

        private boolean testBitsInState(final int mask) {
            return (getState() & mask) != 0;
        }
    }

    static class Claim {
        boolean writer;
        SharedResource resource;
        int count;
        long start;
        volatile boolean granted;

        private Claim(SharedResource resource, boolean writer, int count) {
            this.writer = writer;
            this.count = count;
            this.resource = resource;
            this.start = System.currentTimeMillis();
        }

        public String toString() {
            return String.format("[%s] count=%d writer=%s granted=%s", resource, count, writer, granted);
        }
    }

    final static Map<SessionId, List<Claim>> claims = new ConcurrentHashMap<SessionId, List<Claim>>();

    protected final Persistit _persistit;

    private final Sync _sync = new Sync();

    private volatile int _id;

    /**
     * A counter that increments every time the resource is changed.
     */
    protected AtomicLong _generation = new AtomicLong();

    protected SharedResource(final Persistit persistit) {
        this._persistit = persistit;
    }

    public boolean isAvailable(final boolean writer) {
        return writer ? _sync.isAvailable(_sync.state()) : _sync.isAvailableShared(_sync.state());
    }

    boolean isDirty() {
        return _sync.testBitsInState(DIRTY_MASK);
    }

    public boolean isValid() {
        return _sync.testBitsInState(VALID_MASK);
    }

    public boolean isTemporary() {
        return _sync.testBitsInState(TEMPORARY_MASK);
    }

    boolean isFixed() {
        return _sync.testBitsInState(FIXED_MASK);
    }

    boolean isWriter() {
        return _sync.testBitsInState(WRITER_MASK);
    }

    /**
     * Indicates whether this Thread has a writer claim on this page.
     * 
     * @return <i>true</i> if this Thread has a writer claim on this page.
     */
    boolean isOwnedAsWriterByMe() {
        return (_sync.writerThread() == Thread.currentThread());
    }

    boolean isOwnedAsWriterByOther() {
        final Thread t = _sync.writerThread();
        return t != null && t != Thread.currentThread();
    }

    void changed() {
        assert isOwnedAsWriterByMe();
        final Collection<Thread> threads = _sync.getQueuedThreads();
        _id++;
        for (final Thread t : threads) {
            if (_sync.isQueued(t)) {
                t.interrupt();
            }
        }
    }

    boolean claim(final boolean writer) throws PersistitInterruptedException {
        return claim(writer, DEFAULT_MAX_WAIT_TIME);
    }

    boolean claim(final boolean writer, final long timeout) throws PersistitInterruptedException {
        final Claim claim = recordClaim(writer, 1);
        final boolean result = claim0(writer, timeout);
        if (!result) {
            removeClaim(1);
        }
        claim.granted = true;
        return result;
    }

    boolean claim0(final boolean writer, final long timeout) throws PersistitInterruptedException {
        if (timeout == 0) {
            if (writer) {
                return _sync.tryAcquire(1);
            } else {
                return _sync.tryAcquireShared(1) >= 0;
            }
        } else {
            final int id = _id;
            try {
                if (writer) {
                    if (_sync.tryAcquireNanos(1, timeout * 1000000)) {
                        return true;
                    }
                } else {
                    if (_sync.tryAcquireSharedNanos(1, timeout * 1000000)) {
                        return true;
                    }
                }
            } catch (final InterruptedException e) {
                if (id == _id) {
                    throw new PersistitInterruptedException(e);
                } else {
                    return false;
                }
            }
            return false;
        }
    }

    boolean upgradeClaim() {
        recordClaim(true, 0);
        boolean result = upgradeClaim0();
        if (!result) {
            removeClaim(0);
        }
        return result;
    }

    boolean upgradeClaim0() {
        return _sync.tryUpgrade();
    }

    void release() {
        removeClaim(1);
        _sync.release(1);
    }

    void releaseWriterClaim() {
        removeClaim(0);
        _sync.release(0);
    }

    boolean setDirty() {
        return _sync.setBitsInState(DIRTY_MASK);
    }

    boolean clearDirty() {
        return _sync.clearBitsInState(DIRTY_MASK);
    }

    void setTouched() {
        _sync.setBitsInState(TOUCHED_MASK);
    }

    void clearTouched() {
        _sync.clearBitsInState(TOUCHED_MASK);
    }

    void setFixed() {
        _sync.setBitsInState(FIXED_MASK);
    }

    void clearFixed() {
        _sync.clearBitsInState(FIXED_MASK);
    }

    boolean isTouched() {
        return _sync.testBitsInState(TOUCHED_MASK);
    }

    public long getGeneration() {
        return _generation.get();
    }

    void setTemporary() {
        _sync.setBitsInState(TEMPORARY_MASK);
    }

    void clearTemporary() {
        _sync.clearBitsInState(TEMPORARY_MASK);
    }

    void setValid() {
        _sync.setBitsInState(VALID_MASK);
    }

    void clearValid() {
        _sync.clearBitsInState(VALID_MASK);
    }

    void bumpGeneration() {
        _generation.incrementAndGet();
    }

    public int getStatus() {
        return _sync.state();
    }

    /**
     * Sets bits in the state. This method does not change the bits used by the
     * synchronizer to maintain lock state.
     * 
     * @param mask
     */
    void setStatus(final SharedResource resource) {
        final int mask = resource.getStatus();
        _sync.clearBitsInState(~(WRITER_MASK | CLAIMED_MASK));
        _sync.setBitsInState(mask & ~(WRITER_MASK | CLAIMED_MASK));
    }

    public String getStatusCode() {
        return getStatusCode(getStatus());
    }

    public String getStatusDisplayString() {
        final Thread writerThread = _sync.writerThread();
        final int state = _sync.state();
        if (writerThread == null) {
            return getStatusCode(state);
        } else {
            return getStatusCode(state) + " <" + writerThread.getName() + ">";
        }
    }

    public static String getStatusCode(final int state) {
        // Common cases so we don't have to construct new StringBuilders
        switch (state) {
        case 0:
            return "";
        case VALID_MASK:
            return "v";
        case VALID_MASK | DIRTY_MASK:
            return "vd";
        case VALID_MASK | 1:
            return "vr1";
        case VALID_MASK | WRITER_MASK | 1:
            return "vwr1";
        case VALID_MASK | DIRTY_MASK | 1:
            return "vdr1";
        case VALID_MASK | WRITER_MASK | DIRTY_MASK | 1:
            return "vdwr1";
        default:
            final StringBuilder sb = new StringBuilder(8);

            if ((state & VALID_MASK) != 0) { // TODO chars
                sb.append("v");
            }
            if ((state & DIRTY_MASK) != 0) {
                sb.append("d");
            }
            if ((state & TEMPORARY_MASK) != 0) {
                sb.append("t");
            }
            if ((state & WRITER_MASK) != 0) {
                sb.append("w");
            }
            if ((state & CLAIMED_MASK) != 0) {
                sb.append("r");
                sb.append(state & CLAIMED_MASK);
            }
            return sb.toString();
        }
    }

    public Thread getWriterThread() {
        return _sync.writerThread();
    }

    @Override
    public String toString() {
        return getClass().getSimpleName() + this.hashCode() + " status=" + getStatusDisplayString();
    }

    Claim recordClaim(boolean writer, int count) {
        List<Claim> list = claims.get(_persistit.getSessionId());
        if (list == null) {
            list = new ArrayList<Claim>(4);
            claims.put(_persistit.getSessionId(), list);
        }
        Claim claim = new Claim(this, writer, count);
        synchronized (list) {
            list.add(claim);
        }
        return claim;
    }

    void removeClaim(final int count) {
        List<Claim> list = claims.get(_persistit.getSessionId());
        assert list != null : "list is null";
        int remaining = count;
        synchronized (list) {
            for (int index = list.size(); --index >= 0;) {
                final Claim claim = list.get(index);
                if (claim.resource == this) {
                    int remove = Math.min(claim.count, remaining);
                    claim.count -= remove;
                    remaining -= remove;
                    if (claim.count == 0) {
                        list.remove(index);
                    } else if (count == 0) {
                        claim.writer = false;
                    }
                    if (remaining == 0) {
                        return;
                    }
                }
            }
            assert false : "claim not found";
        }
    }

    private static int armed = 0;

    public static synchronized void reportClaims() {
        if (armed-- < 0) {
            armed = 10;
            System.out.printf("\n\nImpending deadlock:\n");
            System.out.printf("==========================\n");
            for (final Map.Entry<SessionId, List<Claim>> entry : claims.entrySet()) {
                final List<Claim> list = entry.getValue();
                List<Claim> copy = null;
                if (list != null) {
                    synchronized (list) {
                        copy = new ArrayList<Claim>(list);
                    }
                    if (!copy.isEmpty()) {
                        System.out.println();
                        System.out.println(entry.getKey().ownerName());
                        for (final Claim c : copy) {
                            System.out.println(c);
                        }
                        Thread t = entry.getKey().getOwner();
                        if (t != null) {
                            final StackTraceElement[] elements = t.getStackTrace();
                            for (final StackTraceElement e : elements) {
                                System.out.println("  |" + e);
                            }
                        }
                    }
                    
                }
            }
            System.out.printf("==========================\n\n");
        }
    }

    public static void poll() {
        ArrayList<Claim> copy = new ArrayList<Claim>(100);
        long now = System.currentTimeMillis();
        for (final Map.Entry<SessionId, List<Claim>> entry : claims.entrySet()) {
            final List<Claim> list = entry.getValue();
            if (list != null) {
                synchronized (list) {
                    copy.addAll(list);
                }
                if (!copy.isEmpty()) {
                    final Claim claim = copy.get(copy.size() - 1);
                    if (!claim.granted && now - claim.start > ((DEFAULT_MAX_WAIT_TIME * 8L) / 10L)) {
                        reportClaims();
                    }
                    break;
                }
            }
        }
    }
}
